from typing import Any, Generator, Iterable, List, Tuple


def batcher(iterable: Iterable, batch_size: int) -> Generator[List, None, None]:
    """Create a batched generator out of a generator

    :param generator: the iterable that will be batched.
    :param batch_size: The max size of the list
    :yield: a list with batch_size number of items in iterable. If there is a residual
     at the end, the last list will be shorter.
    """
    batch = []
    counter = 0
    for i in iterable:
        batch.append(i)
        counter += 1
        if counter % batch_size == 0:
            yield batch
            batch = []
    if len(batch) > 0:
        yield batch


def add_args_to_iterable(
    iterable: Iterable, *args: Any
) -> Generator[Tuple, None, None]:
    """Create a generator where some args are repeated every time.

    :param iterable: the variable that will change in every iteration
    :yield: a generator that will look like `(iterable[i], args[0], args[1], ...)`
    """
    for i in iterable:
        yield i, *args



import os
import re
from datetime import datetime

from pydicom import Dataset

from cobra_db.model import DicomEntity
from cobra_db.utils import parse_DA_TM_as_datetime

non_alphanumeric = re.compile(r"[^a-zA-Z\d]")
opt = DicomEntity.optional


def get_patient_path(patient_anon_id: str):
    """Obtains a relative folder with the first 12 Bytes (24 chars of the hexstring):
    The hashed PatientID is
    26e3af1f7f22df92d6185a8e15ebdfc0ad089a17ec516377525f42c016f1d5cb
    the folder is
    26e/3af/1f7f22df92d6185a8e/
    This will balance the folder structure given that the hash is almost random."""
    if patient_anon_id is None:
        return "UNK_PatientID"
    assert len(patient_anon_id) == 64, "Length of patient hash is incorrect"
    assert int(patient_anon_id, 16), "Non hex values in string"
    p = patient_anon_id
    return os.path.join(p[0:3], p[3:6], p[6:24])


def get_study_path(patient_anon_id: str, study_date: datetime):
    """Our definition of study is PatientID + StudyDate. This method helps keeping the
    folder structure in the same way"""
    if study_date is None:
        study_date = datetime(1900, 1, 1)
    return os.path.join(
        get_patient_path(patient_anon_id), f'study_{study_date.strftime("%Y%m%d")}'
    )


def get_series_path(ds: Dataset):
    """Get a unique path for a dataset"""

    study_dt = parse_DA_TM_as_datetime(
        opt(ds, "StudyDate", str, "00000000"),
        opt(ds, "StudyTime", str, "000000"),
    )
    series_dt = parse_DA_TM_as_datetime(
        opt(ds, "SeriesDate", str, "00000000"),
        opt(ds, "SeriesTime", str, "000000"),
    )
    if series_dt is None:
        series_key = opt(ds, "SeriesNumber", str, "UNK")
    else:
        series_key = series_dt.strftime("%H%M%S")
    modality = opt(ds, "Modality", str, "UNK")
    description = opt(ds, "SeriesDescription", str, "UNK")
    description = non_alphanumeric.sub("-", description)
    patient_id = ds.get("PatientID", None)
    return os.path.join(
        get_study_path(patient_id, study_dt),
        f"series_{modality}_{series_key}_{description}",
    )


def get_instance_path(ds: Dataset):
    """Get unique path for instance"""
    key = ds.SOPInstanceUID
    return os.path.join(get_series_path(ds), f"{key}.dcm")



import logging
import os
import subprocess
from datetime import datetime
from typing import List


def find_files(path: str, query: str, regex: bool = False):
    """Lazyly use the "find" bash command to recursively get all the files
    that match query on path.

    :param path: path to search
    :param query: query to match
    :param regex: use -regex instead of -name
    :raises FileNotFoundError: when nothing is found
    :yield: absolute paths of files
    """
    flag = "-regex" if regex else "-name"
    process = subprocess.Popen(
        ["find", path, "-type", "f", flag, query],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    for c in iter(process.stdout.readline, b""):
        c = c.decode().rstrip("\n")
        yield c
    if (e := process.stderr.readline().decode().rstrip("\n")) != "":
        raise FileNotFoundError(e)


def find_dcm(path):
    """Lazy iterator to recursively find all files with .dcm extension under a path

    :param path: path to find
    :yield: absolute paths of files
    """
    for c in find_files(path, "*.dcm"):
        yield c


def list_dirs(path: str) -> List[str]:
    """List all directories directly below path

    :param path: path to get directories from
    :return: list of directories as absolute paths
    """
    dirs = [f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))]
    return sorted([os.path.join(path, d) for d in dirs])


def list_files(folder_path: str, extension: str) -> List[str]:
    """List paths of files with <extension> directly inside a folder.

    :param folder_path: the path to the folder
    :param extension: the extension to look for
    :return: list of absolute paths of files
    """
    filenames = []
    extension = f".{extension}" if not extension.startswith(".") else extension
    for f in os.listdir(folder_path):
        if os.path.splitext(f)[1] == extension:
            filenames.append(f)
    return sorted([os.path.join(folder_path, f) for f in filenames])


def parse_DA_TM_as_datetime(DA: str, TM: str) -> datetime:
    """
    Parse DA and TM as datetime
    """
    if DA == "00000000" and TM == "000000":
        return None
    if "." in TM:
        format = "%Y%m%d %H%M%S.%f"
    else:
        format = "%Y%m%d %H%M%S"
    return datetime.strptime(f"{DA} {TM}", format)


def parse_AS_as_int(AS: str) -> int:
    """
    Parse AS as int
    """
    if AS is None:
        logging.warning("AS is empty, returning -1")
        return -1
    if len(AS) > 0:
        if AS[-1] == "Y":
            return int(AS[:-1])
        else:
            raise ValueError(f"AS {AS} is not in years")
    else:
        logging.warning("Not skiping but AS is empty")
        return -1


def parse_DA_as_datetime(DA: str):
    return datetime.strptime(DA, "%Y%m%d")


def intersect_dicts(dicts: List[dict]) -> dict:
    """
    Reads the fist level keys and returns everything that is the same for all dicts.
    """
    ans = dict()
    for k, v in dicts[0].items():
        add_pair = True
        v_s = str(v)
        for d in dicts[1:]:
            # Case key does not exist
            d_v = d.get(k, None)
            if d_v is None:
                add_pair = False
                break
            else:  # Case key exists but values are different
                d_v = str(d_v)
                if d_v != v_s:
                    add_pair = False
                    break
        if add_pair:
            ans[k] = v
    return ans


def intersect_dicts_allow_empty_minority(dicts: List[dict]) -> dict:
    """
    Reads first level keys and returns a dict with all the keys where the values are
    equal.
    If the key does not exist in less than half, but all the other dicts that contain it
    agree on the value, then the key is kept.
    """
    # get a set of all the possible keys
    all_keys = set()
    ans = dict()
    n_dicts = len(dicts)
    majority = (n_dicts // 2) + 1
    for d in dicts:
        for k in d.keys():
            all_keys.add(k)
    for k in sorted(all_keys):
        add_key = True
        # get values for k of all dicts if it exists
        values = [d[k] for d in dicts if d.get(k, None) is not None]
        n_values = len(values)
        if n_values < majority:
            add_key = False
        if n_values > 1:
            str_v0 = str(values[0])
            for v in values[1:]:
                if str_v0 != str(v):
                    add_key = False
                    break
        if n_values == 0:  # in case key existed but had a None
            add_key = False
        if add_key:
            ans[k] = values[0]
    return ans



from typing import Tuple, Union

from bson import ObjectId

Id = Union[ObjectId, str]  # used to make methods accept both str or ObjectId

# Loading and processing MRIs

Point3D = Tuple[int, int, int]  # x, y, z
Point2D = Tuple[int, int]  # x, y
Size3D = Tuple[int, int, int]  # x, y, z
Size2D = Tuple[int, int]  # x, y
Rectangle = Tuple[Point2D, Size2D]  # a starting point and two sides size
Cube = Tuple[Point3D, int]  # a starting point and a side size
Box = Tuple[Point3D, Size3D]  # a starting point and sizes for all dimensions



import logging
import os
from datetime import datetime, timezone
from email.generator import Generator
from getpass import getpass
from typing import List, Tuple, Type, Union
from urllib.parse import quote

from bson import ObjectId
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.errors import DuplicateKeyError

from cobra_db.model import (
    Entity,
    FileSource,
    ImageMetadata,
    Patient,
    RadiologicalSeries,
    RadiologicalStudy,
)
from cobra_db.types import Id


def get_obj_id(_id: Id) -> ObjectId:
    """
    Get ObjectId regardless of the initial type
    """
    return _id if isinstance(_id, ObjectId) else ObjectId(_id)


class Connector:
    """
    Connect and auth to a mongo database.
    """

    def __init__(
        self,
        host: str,
        port: int,
        db_name: str,
        username: str = None,
        password: str = None,
    ):
        """Create a new instance of the Connector.
        Remember that passwords should not be stored in plain text and
        you can use the `get_env_pass` and `get_pass` methods as alternative
        methods for inputting the password.

        :param host: url of the mongodb host. Examples "192.168.1.10",
         "localhost"
        :param port: int specifying the port of your database. The most common
         is 27017
        :param db_name: name of the database you want to access
        :param username: your username, defaults to None
        :param password: your password, defaults to None
        """
        self.host = host
        self.port = port
        self.db_name = db_name
        self.username = username
        self.password = password
        self.connect()

    def connect(self):
        """
        Connect to the mongo database
        """
        self.client = MongoClient(
            self._get_uri(self.host, self.port, self.username, self.password),
            serverSelectionTimeoutMS=5000,
            connectTimeoutMS=5000,
        )

        self.db = self.client[self.db_name]

    @staticmethod
    def _get_uri(host: str, port: str, username=None, password=None, db_name=""):
        if host.endswith("mongodb.net"):
            q_username = quote(username)
            q_password = quote(password)
            return f"mongodb+srv://{q_username}:{q_password}@{host}/{db_name}"
        if username is None:
            return f"mongodb://{host}:{port}/{db_name}"
        elif password is None:
            return f"mongodb://{quote(username)}@{host}:{port}/{db_name}"
        else:
            q_username = quote(username)
            q_password = quote(password)
            return f"mongodb://{q_username}:{q_password}@{host}:{port}/{db_name}"

    def close(self):
        """
        Close the connection to the database
        """
        self.client.close()

    @classmethod
    def get_pass(cls, host: str, port: int, db_name: str, username: str):
        """Create a Connector instance by prompting the user for a password.
        Most useful in jupyter notebooks.
        """
        password = getpass(
            f"Password for {cls._get_uri(host, port, username, db_name)} "
        )
        assert password != "" and password is not None, "Password is empty"
        return cls(host, port, db_name, username, password)

    @classmethod
    def get_env_pass(
        cls,
        host: str,
        port: int,
        db_name: str,
        username: str,
        env_var: str = "MONGOPASS",
    ):
        """Create a Connector instance by getting the password from the environment
        variable defined in env_var. Useful when creating scripts.
        """
        password = os.environ.get(env_var, None)
        assert (
            password is not None
        ), f"Env variable {env_var} is empty, please run \n\
             export {env_var}='mypassword'\n\
            and then try again."
        return cls(host, port, db_name, username, password)

    def __str__(self):
        if self.password is not None:
            safe_password = "".join(["*" for _ in self.password])
        else:
            safe_password = None
        return self._get_uri(
            self.host, self.port, self.username, safe_password, self.db_name
        )


class EntityDao:
    """
    Base class for all entity DAOs.
    Children need to implement a concrete to_dict() method to be able to use
    get_by_id(x, obj=True)
    """

    def __init__(self, connector: Connector, entity_type: Type[Entity]):
        self.connector = connector
        self.entity_type = entity_type
        self.collection: Collection = self._get_collection(self.entity_type)

    def _get_collection(self, entity_type: Type[Entity]) -> Collection:
        """
        Get a collection from the database according to the entity type
        """
        return self.connector.db[entity_type.__name__]

    def get_by_id(self, _id: Id, obj=True) -> Union[Entity, dict]:
        """Find a document by _id."""
        _id = get_obj_id(_id)
        answers = [i for i in self.collection.find({"_id": _id})]
        assert (
            len(answers) == 1
        ), f"{self.__class__.__name__}: The number of docs found for id {_id} is \
            {len(answers)}"
        obj_dict = answers[0]
        if obj:
            if (
                getattr(self.entity_type, "from_dict", None) is None
            ):  # check that the method was implemented
                raise NotImplementedError(
                    f"{self.entity_type} needs to implement a 'from_dict' method"
                )
            return self.entity_type.from_dict(obj_dict)
        else:
            return obj_dict

    def find(self, filter: dict = {}, **kwargs) -> Generator:
        """
        Generate instances for the documents that match the filter.
        This is not a very fast approach, but it is easy to use.

        filter: dict, valid mongodb query to find the desired entities.
        kwargs to pass to pymongo.Collection.find(filter, **kwargs)
        """
        for entity in self.collection.find(filter, **kwargs):
            yield self.entity_type.from_dict(entity)

    def delete_by_id(self, _id: Id):
        """Delete a document from the database by its id.

        :param _id: ObjectId or str version of ObjectId of the document that will be
        deleted.
        """
        _id = get_obj_id(_id)
        result = self.collection.delete_one({"_id": _id})
        assert result.deleted_count == 1

    def insert_one(self, entity_instance: Entity) -> ObjectId:
        """
        Insert an instance into the database.
        """
        _id = self.collection.insert_one(entity_instance.to_dict()).inserted_id
        return _id


class PatientDao(EntityDao):
    """
    Get or insert patient from / to the database
    (Manages access to the Patient collection in MongoDB)
    """

    def __init__(self, connector):
        super().__init__(connector, Patient)
        self._index_exists = False

    def _ensure_index_exists(self):
        """
        Check if the index for anon_id exists
        """
        if not self._index_exists:
            self.collection.create_index("anon_id", unique=True)
            self._index_exists = True

    def anon_id_to_mongo_id(self, anon_id: str) -> ObjectId:
        """
        Get the _id of a patient from its anon_id
        """
        patient = self.collection.find_one({"anon_id": anon_id})
        if patient is None:
            raise IndexError(f"Patient {anon_id} does not exist")
        return patient["_id"]

    def get_patient_by_id(self, _id: Id = None, obj=True) -> Union[Patient, dict]:
        """
        Get a patient from its _id
        """
        _id = get_obj_id(_id)
        patient = self.collection.find_one({"_id": _id})
        if patient is None:
            raise IndexError(f"Patient with _id: {_id} does not exist")
        if obj:
            return Patient(**patient)
        else:
            return patient

    def get_patient_by_anon_id(self, anon_id: str, obj=True) -> Union[Patient, dict]:
        """
        Get a patient from its anon_id
        """
        patient = self.collection.find_one({"anon_id": anon_id})
        if patient is None:
            raise IndexError(f"Patient with anon_id: {anon_id} does not exist")
        if obj:
            return Patient(**patient)
        else:
            return patient

    def get_patient(self, patient: Patient, obj=True) -> Union[Patient, dict]:
        """
        Get a patient if it exists in the database.
        Expect an IndexError if it does not exist.
        """
        if patient._id is not None:
            get_patient = self.get_patient_by_id
            identifier = patient._id
        elif patient.anon_id is not None:
            get_patient = self.get_patient_by_anon_id
            identifier = patient.anon_id
        else:
            raise IndexError(f"Patient {patient} does not have a _id or anon_id")
        patient = get_patient(identifier, obj=False)
        if obj:
            return Patient(**patient)
        else:
            return patient

    def exists(self, patient: Patient) -> Tuple[bool, Union[ObjectId, None]]:
        try:
            return True, self.get_patient(patient, obj=False)["_id"]
        except IndexError:
            return False, None

    def insert_one(self, patient: Patient, skip_duplicates=True):
        """
        Insert a patient into the database.
        raises DuplicateKeyError if the patient already exists and check_unique is False
        """
        if skip_duplicates:
            exists, _id = self.exists(patient)
            if exists:
                logging.info(
                    f"Skipping insertion. Patient already exists in the database with \
                    _id {_id}"
                )
                return _id
        _id = self.collection.insert_one(patient.to_dict()).inserted_id
        self._ensure_index_exists()
        return _id

    def get_by_id(self, _id: Id, obj=True) -> Union[Patient, dict]:
        """redifining only for type hinting"""
        return super().get_by_id(_id, obj)


class StudyDao(EntityDao):
    """
    Get or insert RadiologicalStudy from / to the database
    (Manages access to the RadiologicalStudy collection in MongoDB)
    """

    def __init__(self, connector):
        super().__init__(connector, RadiologicalStudy)
        self.patient_dao = PatientDao(connector)
        self._index_exists = False

    def _ensure_index_exists(self):
        """
        Check if the index for study_uid exists
        """
        if not self._index_exists:
            self.collection.create_index(
                [("dicom_tags.PatientID.Value", 1), ("date", 1)], unique=True
            )
            self._index_exists = True

    def insert_one(self, study: RadiologicalStudy, skip_duplicates=True):
        """
        Insert a study into the database.
        raises DuplicateKeyError if the study already exists and skip_duplicates is
        False
        """
        assert isinstance(
            study, RadiologicalStudy
        ), f"type of study should be RadiologicalStudy, not {type(study)}"
        if skip_duplicates:
            try:
                _id = self.collection.insert_one(study.to_dict()).inserted_id
            except DuplicateKeyError:
                logging.info(
                    f"Skipping insertion. Study already exists in the database with \
                    study_uid {study.study_uid}"
                )
                _id = self.collection.find_one(
                    {
                        "dicom_tags.PatientID.Value": study.get_tag("PatientID"),
                        "date": study.date,
                    }
                )["_id"]
        else:
            _id = self.collection.insert_one(study.to_dict()).inserted_id
        self._ensure_index_exists()
        return _id

    def update_patient_id(self, study_id: Id, patient_id: Id):
        """Once the patient doc is created the series_id of the metadata should be
        updated
        """
        return self.collection.update_one(
            {"_id": get_obj_id(study_id)},
            {
                "$set": {
                    "patient_id": get_obj_id(patient_id),
                    "_metadata.modified": datetime.now(timezone.utc),
                }
            },
        )

    def get_all_ids(
        self, patient: Union[Patient, Id], *modality: str, other_filters: dict = {}
    ) -> List[ObjectId]:
        """
        Given a patient or patient_id, retrieve all studies that match any of
        the specified modality. other_filters allows you to narrow down the query.
        """
        patient_id = (
            get_obj_id(patient._id)
            if isinstance(patient, Patient)
            else get_obj_id(patient)
        )
        query = {
            "patient_id": patient_id,
            "modality": {"$in": modality},
            **other_filters,
        }
        ids = [i["_id"] for i in self.collection.find(query, {"_id": 1})]
        return ids

    def get_by_id(self, _id: Id, obj=True) -> Union[RadiologicalStudy, dict]:
        """Get a RadiologicalStudy by _id"""
        # redifining only for type hinting
        return super().get_by_id(_id, obj)


class SeriesDao(EntityDao):
    """
    Get or insert RadiologicalSeries from / to the database
    (Manages access to the RadiologicalSeries collection in MongoDB)
    """

    def __init__(self, connector):
        super().__init__(connector, RadiologicalSeries)
        self.study_dao = StudyDao(connector)
        self._index_exists = False

    def _ensure_index_exists(self):
        """
        Check if the index for anon_series_id exists
        """
        if not self._index_exists:
            self.collection.create_index(
                [("series_uid", 1), ("study_id", 1)], unique=True
            )
            self._index_exists = True

    def insert_one(
        self,
        series: RadiologicalSeries,
        study: RadiologicalStudy = None,
        skip_duplicates=True,
    ):
        """
        Insert a series into the database first retrieving the study id.
        raises DuplicateKeyError if the series already exists and skip_duplicates is
        False
        """
        if series.study_id is None and study is not None:
            series = self.obtain_study_id(series, study)
        if skip_duplicates:
            try:
                _id = self.collection.insert_one(series.to_dict()).inserted_id
            except DuplicateKeyError:
                logging.info(
                    f"Skipping insertion. Series already exists in the database with \
                        series_uid {series.series_uid}"
                )
                _id = self.collection.find_one({"series_uid": series.series_uid})["_id"]
        else:
            _id = self.collection.insert_one(series.to_dict()).inserted_id
        self._ensure_index_exists()
        return _id

    def update_study_id(self, series_id: Id, study_id: Id):
        """Once the series doc is created the series_id of the metadata should be
        updated
        """
        return self.collection.update_one(
            {"_id": get_obj_id(series_id)},
            {
                "$set": {
                    "study_id": get_obj_id(study_id),
                    "_metadata.modified": datetime.now(timezone.utc),
                }
            },
        )

    def get_all_ids(
        self, study: Union[RadiologicalStudy, Id], other_filters: dict = {}
    ):
        """
        Given a study or study_id, retrieve all series that belong to this study.
        other_filter can be used to narrow down the query.
        """
        study_id = (
            get_obj_id(study._id)
            if isinstance(study, RadiologicalStudy)
            else get_obj_id(study)
        )
        query = {
            "study_id": study_id,
            **other_filters,
        }
        return [i["_id"] for i in self.collection.find(query, {"_id": 1})]

    def get_by_id(self, _id: Id, obj=True) -> Union[RadiologicalSeries, dict]:
        """Get RadiologicalSeries by _id"""
        # redifining only for type hinting
        return super().get_by_id(_id, obj)


class ImageMetadataDao(EntityDao):
    """
    Get or insert an instance of ImageMetadata from / to the database
    (Manages access to the ImageMetadata collection in MongoDB)
    """

    def __init__(self, connector):
        super().__init__(connector, ImageMetadata)
        self.series_dao = SeriesDao(connector)
        self._index_exists = False

    def _ensure_index_exists(self):
        """
        Check if the index for anon_image_id exists
        """
        if not self._index_exists:
            self.collection.create_index(
                [("image_uid", 1), ("series_id", 1)], unique=True
            )
            self._index_exists = True

    def insert_one(self, im_metadata: ImageMetadata):
        _id = self.collection.insert_one(im_metadata.to_dict()).inserted_id
        return _id

    def update_series_id(self, im_meta_id: Id, series_id: Id):
        """Once the series doc is created the series_id of the metadata should be
        updated
        """
        return self.collection.update_one(
            {"_id": get_obj_id(im_meta_id)},
            {
                "$set": {
                    "series_id": get_obj_id(series_id),
                    "_metadata.modified": datetime.now(timezone.utc),
                }
            },
        )

    def update_study_id(self, im_meta_id: Id, study_id: Id):
        """Once the series doc is created the series_id of the metadata should be
        updated
        """
        return self.collection.update_one(
            {"_id": get_obj_id(im_meta_id)},
            {
                "$set": {
                    "study_id": get_obj_id(study_id),
                    "_metadata.modified": datetime.now(timezone.utc),
                }
            },
        )

    def get_all_ids(
        self, series: Union[RadiologicalSeries, Id], other_filters: dict = {}
    ):
        """Given a series or series_id, retrieve all images that belong to this series.
        other_filter can be used to narrow down the query.
        """
        study_id = (
            get_obj_id(series._id)
            if isinstance(series, RadiologicalSeries)
            else get_obj_id(series)
        )
        query = {
            "series_id": study_id,
            **other_filters,
        }
        return [i["_id"] for i in self.collection.find(query, {"_id": 1})]

    def get_by_id(self, _id: Id, obj=True) -> Union[ImageMetadata, dict]:
        """Get an ImageMetadata by _id"""
        # redifining only for type hinting
        return super().get_by_id(_id, obj)

    def add_aka(self, im_meta_id: Id, file_source: FileSource):
        """Adds an AKA file source. This is useful when a pseudonymized version of the
        file is saved somewhere else.
        :param im_meta_id: The id of the document that will be updated
        :param file_source: The AKA filesource
        """
        return self.collection.update_one(
            {"_id": get_obj_id(im_meta_id)},
            {
                "$set": {
                    "_metadata.modified": datetime.now(timezone.utc),
                },
                "$push": {
                    "aka_file_sources": file_source.to_dict(),
                },
            },
        )



from typing import Callable

from pydicom.datadict import keyword_for_tag, tag_for_keyword


class DatasetMod:
    """A functional class to replace the tags with keywords and viceversa in the output
    of pydicom.dataset.Dataset.to_json_dict
    """

    @staticmethod
    def tag_for_keyword_wrapper(keyword: str) -> str:
        """wrapper to get strings instead of ints"""
        if keyword.startswith("60") or keyword.startswith(
            "50"
        ):  # return the keyword as is
            return keyword
        tag_int = tag_for_keyword(keyword)
        if tag_int is not None:
            return f"{tag_int:08x}"
        else:  # There is no Keyword for the tag
            return keyword

    @staticmethod
    def keyword_for_tag_wrapper(tag: str) -> str:
        """Ignore translation of Repeating Groups. See
        https://pydicom.github.io/pydicom/dev/old/working_with_overlays.html
        """
        keyword = keyword_for_tag(tag)
        if (
            tag.startswith("60")
            or tag.startswith("50")
            or keyword.startswith("Overlay")
        ):
            return tag
        if keyword == "":  # if the tag is unknown, then leave the tag
            return tag
        return keyword

    @staticmethod
    def _replace_dict(d: dict, replace_func: Callable) -> dict:
        """d: a posibly nested dict where all tag keys are converted into keywords"""
        ans = dict()
        for old_k, v in d.items():
            new_k = replace_func(old_k)
            # check if value is nested
            if v["vr"] == "SQ":
                values = v.pop("Value", None)
                if values is not None:
                    v["Value"] = [
                        DatasetMod._replace_dict(seq_obj, replace_func)
                        for seq_obj in values
                    ]
            ans[new_k] = v
        return ans

    @staticmethod
    def tags_to_keywords(d: dict) -> dict:
        return DatasetMod._replace_dict(d, DatasetMod.keyword_for_tag_wrapper)

    @staticmethod
    def keywords_to_tags(d: dict) -> dict:
        return DatasetMod._replace_dict(d, DatasetMod.tag_for_keyword_wrapper)



import sys
from argparse import ArgumentParser
from typing import List, Tuple

from cobra_db.scripts.stage_2_group_patients import main as group_patients
from cobra_db.scripts.stage_2_group_patients import (
    main_multiproc as group_patients_multiproc,
)
from cobra_db.scripts.stage_2_group_series import main as group_series
from cobra_db.scripts.stage_2_group_series import (
    main_multiproc as group_series_multiproc,
)
from cobra_db.scripts.stage_2_group_studies import main as group_studies
from cobra_db.scripts.stage_2_group_studies import (
    main_multiproc as group_studies_multiproc,
)


def parse_arguments(raw_args: List[str]) -> Tuple[str, dict, int]:
    """Convert the console arguments into python variables.

    :param raw_args: the list of commands (usually sys.argv[1:])
    :return: connector_kwargs, n_proc
    """
    parser = ArgumentParser(
        description="Once the ImageMetadata collection is populated, group the images \
        into series, studies and patients."
    )
    parser.add_argument("-ho", "--host", help="MongoDB host", default="127.0.0.1")
    parser.add_argument("-p", "--port", help="MongoDB port", default=27017, type=int)
    parser.add_argument("-u", "--username", help="MongoDB username", default=None)
    parser.add_argument(
        "-db", "--database_name", help="MongoDB database", required=True
    )
    parser.add_argument(
        "-n", "--num_processes", help="Number of processes to use", default=1, type=int
    )
    parser.add_argument(
        "--project_name",
        help="The name of the project that gets stored in the _metadata field",
    )

    args = parser.parse_args(raw_args)
    connector_kwargs = {
        "host": args.host,
        "port": args.port,
        "username": args.username,
        "db_name": args.database_name,
    }
    return connector_kwargs, args.num_processes, args.project_name


def main(connector_kwargs: dict, n_proc: int, project_name: str):
    if n_proc == 1:
        group_series(connector_kwargs, project_name)
        group_studies(connector_kwargs, project_name)
        group_patients(connector_kwargs)  # patients dont have a project.
    else:
        group_series_multiproc(connector_kwargs, n_proc, project_name)
        group_studies_multiproc(connector_kwargs, n_proc, project_name)
        group_patients_multiproc(connector_kwargs, n_proc)


def cli():
    connector_kwargs, n_proc, project_name = parse_arguments(sys.argv[1:])
    main(connector_kwargs, n_proc, project_name)



from enum import Enum, IntFlag
from typing import Literal


class ByEnumName:
    """Marker to specify that this enum is stored by it's name in the db"""


class StrEnum(str, Enum):
    pass


class ProjectName(StrEnum):
    """Project name used in metadata as enum to keep it tidy"""

    VAIB = "vaib"
    MAMMOAI = "mammoai"


class Modality(StrEnum):
    """Modality of the image.
    See https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.3.html
    sect_C.7.3.1.1.1
    """

    MR = "MR"  # Magnetic Resonance
    MG = "MG"  # Mammography
    CR = "CR"  # Computed Radiography. Sometimes used instead of mammography MG
    CT = "CT"  # Computed Tomography
    SR = "SR"  # Structured Reporting
    US = "US"  # Ultrasound
    DX = "DX"  # Digital Radiography
    OTHER = "OT"  # Other
    PT = "PT"


class DataFormat(StrEnum):
    """
    Format of an image
    """

    DICOM = "dicom"
    JPEG = "jpeg"
    PNG = "png"
    TIFF = "tiff"


class Laterality(StrEnum):
    """Laterality of the image (0020,0060). Adding "bilateral" for MRIs that contain
    both breasts.
    """

    LEFT = "left"
    RIGHT = "right"
    BILATERAL = "bilateral"

    @classmethod
    def from_image_laterality(cls, image_laterality: str):
        if isinstance(image_laterality, str):
            if image_laterality.lower() == "l":
                return cls("left")
            elif image_laterality.lower() == "r":
                return cls("right")
        return None

    @classmethod
    def from_nkbc_a_pat_sida(cls, value: Literal[1, 2, "1", "2"]):
        """Map the description of a_pat_sida to left or right according to

        :param value: value coming from the NKBC database/csv. Literal[1, 2]
        :return: instance of this class
        """
        if value is None:
            return None
        if type(value) == str:
            value = int(value)
        if value == 1:
            return cls("right")
        if value == 2:
            return cls("left")
        if value == 3:
            return cls("bilateral")
        raise ValueError(
            f"Value {value} for a_pat_sida could not be mapped to 'right' or 'left'"
        )


class View(StrEnum):
    """
    View of the image.
    """

    AX = "axial"
    SAG = "sagittal"
    COR = "coronal"
    MLO = "mediolateral_oblique"
    CC = "craniocaudal"

    @classmethod
    def from_view_position(cls, view_position: str):
        if isinstance(view_position, str):
            if view_position in ["AX", "SAG", "COR", "MLO", "CC"]:
                return getattr(cls, view_position)
        return None


class PregnancyStatus(StrEnum):
    """
    Pregnancy status of the patient. (0010,21C0).
    """

    NOT_PREGNANT = "0001"
    POSSIBLY_PREGNANT = "0002"
    DEFINETLY_PREGNANT = "0003"
    UNKNOWN = "0004"


class PatientPosition(StrEnum):
    """Patient Position
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.3.html
    sect_C.7.3.1.1.2
    """

    HFP = "HFP"  # "Head First-Prone"
    HFS = "HFS"  # "Head First-Supine"
    HFDR = "HFDR"  # "Head First-Decubitus Right"
    HFDL = "HFDL"  # "Head First-Decubitus Left"
    FFDR = "FFDR"  # "Feet First-Decubitus Right"
    FFDL = "FFDL"  # "Feet First-Decubitus Left"
    FFP = "FFP"  # "Feet First-Prone"
    FFS = "FFS"  # "Feet First-Supine"
    LFP = "LFP"  # "Left First-Prone"
    LFS = "LFS"  # "Left First-Supine"
    RFP = "RFP"  # "Right First-Prone"
    RFS = "RFS"  # "Right First-Supi"
    AFDR = "AFDR"  # "Anterior First-Decubitus Right"
    AFDL = "AFDL"  # "Anterior First-Decubitus Left"
    PFDR = "PFDR"  # "Posterior First-Decubitus Right"
    PFDL = "PFDL"  # "Posterior First-Decubitus Left"


class PersonalNumberValidity(IntFlag):
    """Describes how valid a Swedish Personal Number is.
    As a binary set of flags where 1 means valid.
    0b |0        |0    |0      |0
       |checksum |date |length |chars
    In the end all the flags are interpreted as an integer. For example:
    3 would be 0b0011 which means valid chars and length.
    6 would be 0b0110 which means valid length and date.
    7 would be 0b0111 which means only the checksum is not valid
    And so on for any number between 0 and 15
    """

    NOT_VALID = 0
    CHARS = 1  # only 0-9, a-z, or A-Z characters. No other dashes dots slashes etc.
    LENGTH = 2  # 12 characters YYYYMMDDXXXX (after removing other characters that are
    # not 0-9 a-z A-Z)
    DATE = 4  # A date can be parsed (when using only YYMMDD)
    CHECKSUM = 8  # The checksum is correct
    VALID = 15  # everything is correct



# read version from installed package
from importlib.metadata import version

__version__ = version("cobra_db")

from cobra_db.model import ImageMetadata, Patient, RadiologicalSeries, RadiologicalStudy
from cobra_db.mongo_dao import (
    Connector,
    ImageMetadataDao,
    PatientDao,
    SeriesDao,
    StudyDao,
)



import logging
import os
from copy import deepcopy
from datetime import datetime
from typing import List, Union

import pydicom
from deid.config import DeidRecipe
from deid.dicom.parser import DicomParser
from deid.logger import bot

from cobra_db.encrypt import Hasher
from cobra_db.utils import parse_AS_as_int, parse_DA_TM_as_datetime

base_recipe_path = os.path.join(os.path.dirname(__file__), "deid_recipe.txt")
mr_recipe_path = os.path.join(os.path.dirname(__file__), "deid_recipe_mr.txt")

deid_logging_levels = dict(
    ABORT=-5,
    FLAG=-4,
    ERROR=-3,
    WARNING=-2,
    LOG=-1,
    INFO=1,
    CUSTOM=1,
    QUIET=0,
    VERBOSE=2,
    VERBOSE2=3,
    VERBOSE3=4,
    DEBUG=5,
)


class Deider:
    def __init__(
        self,
        hasher_secret_salt: str,
        recipe_path: Union[str, List] = None,
        logging_level: str = "ERROR",
    ):
        """Deidentify datasets according to vaib recipe

        :param recipe_path: path to the deid recipe
        :param hasher_secret_salt: salt for hashing
        """
        bot.level = deid_logging_levels[logging_level]
        if recipe_path is None:
            logging.warning(f"DeidDataset using default recipe {base_recipe_path}")
            recipe_path = base_recipe_path
        if type(recipe_path) == list:
            for r in recipe_path:
                assert os.path.exists(r), f"Invalid recipe_path: {r}"
        else:
            assert os.path.exists(recipe_path), f"Invalid recipe_path: {recipe_path}"
        self.recipe = DeidRecipe(recipe_path)
        self.hasher = Hasher(hasher_secret_salt)

    def pseudonymize(self, dataset: pydicom.Dataset) -> pydicom.Dataset:
        """Pseudonymize a single dicom dataset

        :param dataset: dataset that will be pseudonymized
        :returns: pseudonymized dataset
        """
        dataset = deepcopy(dataset)
        parser = DicomParser(dataset, self.recipe)
        parser.define("replace_name", self._replace_name)
        parser.define("hash_func", self._deid_hash_func)
        parser.define("remove_day", self._remove_day)
        parser.define("round_AS_to_nearest_5y", self._round_AS_to_nearest_5y)
        parser.define("round_DS_to_nearest_5", self._round_DS_to_nearest_5)
        parser.define("round_DS_to_nearest_0_05", self._round_DS_to_nearest_0_05)
        parser.parse(strip_sequences=False, remove_private=True)
        return parser.dicom

    @staticmethod
    def _remove_day(item, value, field, dicom):
        """Removes the day from a DT field in the deid framework"""
        date = field.element.value
        if date == "":
            return ""
        dt = datetime.strptime(date, "%Y%m%d")
        return dt.strftime("%Y%m01")

    @staticmethod
    def _replace_name(item, value, field, dicom):
        sex = dicom.get("PatientSex")
        sex = {"F": "Female", "M": "Male", "O": "Other", "": "Unk"}[sex]
        age = Deider._round_to_nearest(parse_AS_as_int(dicom.get("PatientAge")), 5)
        return f"{sex} {age:03d}Y {dicom.get('Modality')}"

    @staticmethod
    def _round_to_nearest(value, interval):
        """Rounds value to closest multiple of interval"""
        return interval * round(value / interval)

    @staticmethod
    def _round_AS_to_nearest_5y(item, value, field, dicom):
        """Rounds age(AS) field to 5 year intervals in the deid framework"""
        value = field.element.value
        # if age is empty, try to calculate it from the StudyDate and PatientBirthDate
        if value == "" or value is None:
            try:
                study_date = dicom.get("StudyDate")
                study_date = parse_DA_TM_as_datetime(DA=study_date, TM="000000")
                birth_date = dicom.get("PatientBirthDate")
                birth_date = parse_DA_TM_as_datetime(DA=birth_date, TM="000000")
                age = (study_date - birth_date).days / 365
            except Exception as e:
                logging.error(e)
                return ""
        else:
            age = parse_AS_as_int(field.element.value)
        return f"{Deider._round_to_nearest(age, 5):03d}Y"

    @staticmethod
    def _round_DS_to_nearest_5(item, value, field, dicom):
        """Rounds age(AS) field to 5 year intervals in the deid framework"""
        value = field.element.value
        if value is None:
            value = -1
        return f"{Deider._round_to_nearest(float(value), 5)}"

    @staticmethod
    def _round_DS_to_nearest_0_05(item, value, field, dicom) -> str:
        """Rounds field.element.value to increments of 0.05"""
        value = field.element.value
        if value is None:
            value = -1
        return f"{Deider._round_to_nearest(float(value), 0.05):.02f}"

    def _deid_hash_func(self, item, value, field, dicom) -> str:
        """Performs self.hash to field.element.value"""
        val = field.element.value
        return self.hasher.hash(str(val))



"""
usage: cobra_ingest [-h]
 -dn DRIVE_NAMES [DRIVE_NAMES ...]
 -dp DRIVE_PATHS [DRIVE_PATHS ...]
 [-ho HOST] [-p PORT] [-u USERNAME] -db DATABASE_NAME [-n NUM_PROCESSES]
 [--project_name PROJECT_NAME]

Recursively ingest a set of directories of dicom files into a mongo database.

optional arguments:
  -h, --help            show this help message and exit
  -dn DRIVE_NAMES [DRIVE_NAMES ...], --drive_names DRIVE_NAMES [DRIVE_NAMES ...]
                        the name of the drives to ingest
  -dp DRIVE_PATHS [DRIVE_PATHS ...], --drive_paths DRIVE_PATHS [DRIVE_PATHS ...]
                        the path where the drives are mounted (in the same order as
                        drive_names)
  -ho HOST, --host HOST
                        MongoDB host
  -p PORT, --port PORT  MongoDB port
  -u USERNAME, --username USERNAME
                        MongoDB username
  -db DATABASE_NAME, --database_name DATABASE_NAME
                        MongoDB database
  -n NUM_PROCESSES, --num_processes NUM_PROCESSES
                        Number of processes to use. It runs single threaded by default
                        so that it can be easily debugged
  --project_name PROJECT_NAME
                        A project name that will be stored in the _metadata field of the
                         created ImageMetadata docs
"""

import logging
import os
import sys
from argparse import ArgumentParser
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Pool
from typing import List, Tuple

import pydicom
from tqdm import tqdm

from cobra_db.model import ImageMetadata
from cobra_db.mongo_dao import Connector, ImageMetadataDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher
from cobra_db.utils import find_dcm


def parse_arguments(raw_args: List[str]) -> Tuple[str, dict, int, str]:
    """Convert the console arguments into python variables.

    :param raw_args: the list of commands (usually sys.argv[1:])
    :return: mount_paths, connector_kwargs, n_proc
    """
    parser = ArgumentParser(
        description="Recursively ingest a set of directories \
            of dicom files into a mongo database."
    )
    parser.add_argument(
        "-dn",
        "--drive_names",
        nargs="+",
        help="the name of the drives to ingest",
        required=True,
    )
    parser.add_argument(
        "-dp",
        "--drive_paths",
        nargs="+",
        help="the path where the drives are mounted (in the same order as drive_names)",
        required=True,
    )
    parser.add_argument("-ho", "--host", help="MongoDB host", default="127.0.0.1")
    parser.add_argument("-p", "--port", help="MongoDB port", default=27027, type=int)
    parser.add_argument("-u", "--username", help="MongoDB username", default=None)
    parser.add_argument(
        "-db", "--database_name", help="MongoDB database", required=True
    )
    parser.add_argument(
        "-n",
        "--num_processes",
        help="Number of processes to use. \
        It runs single threaded by default so that it can be easily debugged",
        default=1,
        type=int,
    ),
    parser.add_argument(
        "--project_name",
        help="A project name that will be stored in the _metadata field \
            of the created ImageMetadata docs",
    )

    args = parser.parse_args(raw_args)
    assert len(args.drive_paths) == len(args.drive_names)
    mount_paths = {n: d for n, d in zip(args.drive_names, args.drive_paths)}
    connector_kwargs = {
        "host": args.host,
        "port": args.port,
        "username": args.username,
        "db_name": args.database_name,
    }
    return mount_paths, connector_kwargs, args.num_processes, args.project_name


def ingest_metadata(
    filepath: str, im_dao: ImageMetadataDao, mount_paths: dict, project_name
) -> int:
    """Ingest a single dicom file's metadata into the database.
    return: 1 if it was ingested properly.
    """
    try:
        ds = pydicom.dcmread(filepath, stop_before_pixels=True)
        # private tags are causing problems because they are
        # not standard. We are not indexing them.
        ds.remove_private_tags()
        image_metadata = ImageMetadata.from_dataset(ds, mount_paths)
        image_metadata._metadata.project_name = project_name
        im_dao.insert_one(image_metadata)
        return 1
    # except OperationFailure as e:
    #     raise e
    # except FileNotFoundError as e:
    #     raise e
    except Exception as e:
        logging.error(f"{e} - filepath:{filepath}")
    return 0


def process_files_batch(args) -> Tuple[int, int]:
    """Process a batch of files.

    :param args: Single argument tuple of (filepaths, im_dao, mount_paths).
     It gives the ability to use multiprocessing.
    :return: number of seen files and number of correctly processed files
    """
    filepaths, connector_kwargs, mount_paths, project_name = args
    connector = Connector.get_env_pass(**connector_kwargs)
    im_dao = ImageMetadataDao(connector)
    seen = len(filepaths)
    ingested = 0
    for filepath in filepaths:
        ingested += ingest_metadata(filepath, im_dao, mount_paths, project_name)
    return seen, ingested


def multiproc_drive(args):
    mount_paths, connector_kwargs, n_proc, drive_name, drive_path, project, pos = args
    logging.warning(f"Ingesting {drive_name} on path {drive_path}")
    print(f"Ingesting {drive_name}")
    pbar_seen = tqdm(
        smoothing=0, desc=f"{drive_name} Seen", unit="dcm", position=pos * 2
    )
    pbar_ingested = tqdm(
        smoothing=0, desc=f"{drive_name} Ingested", unit="dcm", position=pos * 2 + 1
    )
    filepaths = find_dcm(drive_path)
    args_gen = add_args_to_iterable(
        batcher(filepaths, 1000), connector_kwargs, mount_paths, project
    )
    pool = Pool(n_proc)
    total_ingested = 0
    for seen, ingested in pool.imap_unordered(process_files_batch, args_gen):
        pbar_seen.update(seen)
        pbar_ingested.update(ingested)
        total_ingested += ingested
    return total_ingested


def multiproc(
    mount_paths: dict, connector_kwargs: dict, n_proc: int, project_name: str
):
    # ingest each drive
    drive_tasks = []
    for i, (drive_name, drive_path) in enumerate(mount_paths.items()):
        drive_tasks.append(
            [
                mount_paths,
                connector_kwargs,
                n_proc,
                drive_name,
                drive_path,
                project_name,
                i,
            ]
        )
    total_ingested = 0

    with ProcessPoolExecutor(len(drive_tasks)) as executor:
        for ingested in executor.map(multiproc_drive, drive_tasks):
            total_ingested += ingested

    print(f"Finished ingesting {len(mount_paths)} drive(s)")
    return total_ingested


def single_proc(mount_paths: dict, connector_kwargs: dict, project_name: str):
    """Run the same thing as multiproc but single threaded

    :param mount_paths: Where the drives are
    :param connector_kwargs: How to connect to the db
    :param project_name: name of the project that the images belong to
    """
    for drive_name, drive_path in mount_paths.items():
        logging.info(f"Ingesting {drive_name}")
        print(f"Ingesting {drive_name}")
        pbar_seen = tqdm(smoothing=0, desc="Seen", unit="dcm")
        pbar_ingested = tqdm(smoothing=0, desc="Ingested", unit="dcm")
        filepaths = find_dcm(drive_path)
        args_gen = add_args_to_iterable(
            batcher(filepaths, 10), connector_kwargs, mount_paths, project_name
        )
        for args in args_gen:
            seen, ingested = process_files_batch(args)
            pbar_seen.update(seen)
            pbar_ingested.update(ingested)
    print(f"Finished ingesting {len(mount_paths)} drives")


def main(mount_paths, connector_kwargs, n_proc, project_name):
    # configure logging
    # now = datetime.now().strftime("%Y%m%d_%H%M")
    # log_path = f"stage1_{now}.log"
    logging.basicConfig(
        # filename=log_path,
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    logging.warning(
        f"Starting Stage1 with {len(mount_paths)} drives and {n_proc} processes per drive"
    )

    # quick check to make sure we have the drives connected
    for drive_name, mount_path in mount_paths.items():
        assert os.path.exists(
            mount_path
        ), f"Drive {drive_name} does not exist in {mount_path}"

    # ingest everything
    if n_proc == 1:
        total_ingested = single_proc(mount_paths, connector_kwargs, project_name)
    else:
        total_ingested = multiproc(mount_paths, connector_kwargs, n_proc, project_name)

    # Health check afterwards
    im_dao = ImageMetadataDao(Connector.get_env_pass(**connector_kwargs))
    total_images = im_dao.collection.count_documents({})
    print(total_images)

    logging.warning(
        f"Total ingested = {total_ingested},\n\
        total images in {str(im_dao.connector)} = {total_images}"
    )
    logging.warning(
        "Creating collection indices for SOPInstanceUID, \
        SeriesInstanceUID, StudyInstanceUID. \
        See https://www.mongodb.com/docs/manual/indexes/"
    )
    for i in ["SOPInstanceUID", "SeriesInstanceUID", "StudyInstanceUID"]:
        try:
            im_dao.collection.create_index(f"dicom_tags.{i}.Value")
        except Exception as e:
            print(e)
    with_SOPInstanceUID = im_dao.collection.count_documents(
        {"dicom_tags.SOPInstanceUID.Value": {"$exists": True}}
    )
    with_SeriesInstanceUID = im_dao.collection.count_documents(
        {"dicom_tags.SOPInstanceUID.Value": {"$exists": True}}
    )
    with_StudyInstanceUID = im_dao.collection.count_documents(
        {"dicom_tags.SOPInstanceUID.Value": {"$exists": True}}
    )
    logging.warning(
        f"with_SOPInstanceUID: {with_SOPInstanceUID},\n\
        with_SeriesInstanceUID:{with_SeriesInstanceUID},\n\
        with_StudyInstanceUID:{with_StudyInstanceUID}"
    )

    counts = list(
        im_dao.collection.aggregate(
            [
                {
                    "$group": {
                        "_id": "$dicom_tags.SOPInstanceUID.Value",
                        "image_count": {"$sum": 1},
                        "series_instance_uids": {
                            "$addToSet": {
                                "$first": "$dicom_tags.SeriesInstanceUID.Value"
                            }
                        },
                        "study_instance_uids": {
                            "$addToSet": {
                                "$first": "$dicom_tags.StudyInstanceUID.Value"
                            }
                        },
                    }
                },
                {
                    "$facet": {  # split the pipeline to count different things
                        "unique_SOPInstanceUID": [{"$count": "n"}],
                        "instances_with_more_than_1_image": [
                            {"$match": {"image_count": {"$gt": 1}}},
                            {"$count": "n"},
                        ],
                        "instances_with_more_than_1_series_uid": [
                            {"$match": {"series_instance_uids.1": {"$exists": True}}},
                            {"$count": "n"},
                        ],
                        "instances_with_more_than_1_study_uid": [
                            {"$match": {"study_instance_uids.1": {"$exists": True}}},
                            {"$count": "n"},
                        ],
                        "instances_with_more_than_1_series_and_study_uid": [
                            {"$match": {"study_instance_uids.1": {"$exists": True}}},
                            {"$count": "n"},
                        ],
                    }
                },
                {
                    "$project": {
                        "unique_SOPInstanceUID": {"$first": "$unique_SOPInstanceUID.n"},
                        "instances_with_more_than_1_image": {
                            "$first": "$instances_with_more_than_1_image.n"
                        },
                        "instances_with_more_than_1_series_uid": {
                            "$first": "$instances_with_more_than_1_series_uid.n"
                        },
                        "instances_with_more_than_1_study_uid": {
                            "$first": "$instances_with_more_than_1_study_uid.n"
                        },
                        "instances_with_more_than_1_series_and_study_uid": {
                            "$first": "$instances_with_more_than_1_series_and_study_uid.n"
                        },
                    }
                },
            ],
            allowDiskUse=True,
        )
    )[0]
    logging.warning(f"{counts}")


def cli():
    mount_paths, connector_kwargs, n_proc, project_name = parse_arguments(
        raw_args=sys.argv[1:]
    )
    main(mount_paths, connector_kwargs, n_proc, project_name)


if __name__ == "__main__":
    cli()



# In stage 2 we create the hierarchy of files from the metadata.
# First, we group files by SeriesInstanceUID

import logging
from multiprocessing import Pool
from typing import List, Tuple

from pydicom.dataset import Dataset
from tqdm import tqdm

from cobra_db.dataset_mod import DatasetMod
from cobra_db.model import RadiologicalSeries
from cobra_db.mongo_dao import Connector, ImageMetadataDao, SeriesDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher
from cobra_db.utils import intersect_dicts_allow_empty_minority


def proto_series_generator(connector_kwargs):
    query = [
        {"$match": {"dicom_tags.SeriesInstanceUID.Value": {"$exists": True}}},
        # {
        #     "$limit": 50000
        # },
        {
            "$group": {
                "_id": {"$first": "$dicom_tags.SeriesInstanceUID.Value"},
                "instance_ids": {"$push": "$_id"},
                "study_uids": {"$first": "$dicom_tags.StudyInstanceUID.Value"},
                "patient_anon_ids": {"$first": "$dicom_tags.PatientID.Value"},
            }
        },
    ]
    connector = Connector.get_env_pass(**connector_kwargs)
    dao = ImageMetadataDao(connector)
    series_cursor = dao.collection.aggregate(query, allowDiskUse=True)
    for i in series_cursor:
        yield i


def process_series(
    proto_series: dict,
    im_dao: ImageMetadataDao,
    series_dao: SeriesDao,
    project_name: str,
):
    dicom_tags = [
        im_dao.get_by_id(i, obj=False)["dicom_tags"]
        for i in proto_series["instance_ids"]
    ]
    proto_series["dicom_tags"] = intersect_dicts_allow_empty_minority(dicom_tags)
    shared_ds_dict = DatasetMod.keywords_to_tags(proto_series["dicom_tags"])
    shared_ds = Dataset.from_json(shared_ds_dict)
    series = RadiologicalSeries.from_dataset(shared_ds)
    series._metadata.project_name = project_name
    series.image_count = len(proto_series["instance_ids"])  # tags are not trustable
    series_id = series_dao.insert_one(series)
    for i in proto_series["instance_ids"]:
        im_dao.update_series_id(i, series_id)


def process_series_batch(args: Tuple[List[dict], dict]):
    batch, connector_kwargs, project_name = args
    connector = Connector.get_env_pass(**connector_kwargs)
    im_meta_dao = ImageMetadataDao(connector)
    series_dao = SeriesDao(connector)
    counter = 0
    for proto_series in batch:
        try:
            process_series(proto_series, im_meta_dao, series_dao, project_name)
        except Exception as e:
            logging.error(f"SeriesUID {proto_series['_id']} - {e}")
        counter += 1
    connector.client.close()
    return counter


def sanity_check(connector_kwargs):
    connector = Connector.get_env_pass(**connector_kwargs)
    im_dao = ImageMetadataDao(connector)
    total_docs = im_dao.collection.count_documents({})
    docs_missing_series_uid = im_dao.collection.count_documents(
        {"dicom_tags.SeriesInstanceUID.Value": {"$exists": False}}
    )
    logging.warning(
        f"Total images: {total_docs}. Missing SeriesInstanceUID tag:\
        {docs_missing_series_uid} \
        ({round(docs_missing_series_uid*100/total_docs, 4)}%)"
    )


def main_multiproc(connector_kwargs, n_proc, project_name):
    # sanity check before starting
    sanity_check(connector_kwargs)
    pool = Pool(n_proc)
    iterable = batcher(proto_series_generator(connector_kwargs), 5)
    iterable = add_args_to_iterable(iterable, connector_kwargs, project_name)
    pbar = tqdm(desc="series", smoothing=0.005)
    for n in pool.imap_unordered(process_series_batch, iterable, chunksize=1):
        pbar.update(n)


def main(connector_kwargs, project_name):
    sanity_check(connector_kwargs)
    connector = Connector.get_env_pass(**connector_kwargs)
    series_dao = SeriesDao(connector)
    im_dao = ImageMetadataDao(connector)
    iterable = proto_series_generator(connector_kwargs)
    pbar = tqdm(desc="series", smoothing=0.005)
    for proto_series in iterable:
        try:
            process_series(proto_series, im_dao, series_dao, project_name)
            pbar.update()
        except Exception as e:
            logging.error(f"SeriesUID {proto_series['_id']} - {e}")


# if __name__ == "__main__":
#      # configure logging
#     now = datetime.now().strftime("%Y%m%d_%H%M")
#     log_path = f"stage2_group_series_{now}.log"
#     logging.basicConfig(
#         filename=log_path,
#         level=logging.WARNING,
#         format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
#     )
#     connector_kwargs = dict(
#         host="127.0.0.1",
#         port=27027,
#         username="fercos",
#         db_name="coma_original"
#     )
#     main_multiproc(connector_kwargs)



import logging
import os
import sys
from argparse import ArgumentParser, Namespace
from multiprocessing import Pool
from time import sleep
from typing import List, Union

import pydicom
from bson import ObjectId
from pyaml_env import parse_config
from pydicom.errors import BytesLengthException
from pymongo.errors import DuplicateKeyError
from tqdm import tqdm

from cobra_db.dataset_mod import DatasetMod
from cobra_db.deid import Deider, base_recipe_path, mr_recipe_path
from cobra_db.filesystem import get_instance_path
from cobra_db.model import FileSource, ImageMetadata
from cobra_db.mongo_dao import Connector, ImageMetadataDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher


def parse_arguments(raw_args: List[str]) -> str:
    """Parse the arguments for the pseudonymization process

    :param raw_args: arguments from sys.argv[1:]
    :return: a namespace with the required cfg
    """
    parser = ArgumentParser(
        "Pseudonymize ImageMetadata",
        description="""
Given an ImageMetadata collection with real data, create a second database with the
pseudonymized version. The files will be pseudonymized and stored according to the
suggested filesystem paths. Additionally, docs that are correctly pseudonymized will
be updated with a `deid_file_source` field that allows to check where the data went.""",
    )

    parser.add_argument(
        "config_path",
        type=str,
        help="the config file to be used.\
        For an example of the config file, please read the documentation how-to guide",
    )

    args = parser.parse_args(raw_args)
    return Namespace(**parse_config(args.config_path, default_value=None))


def im_meta_generator(src_mongo: dict, query: dict = {}):
    """Generator of im_meta dicts

    :param src_mongo: kwargs for the Connector
    :param query: query to select subset of the ImageMetadata collection
    :yield: dict with image metadata.
    """

    connector = Connector(**src_mongo)
    im_dao = ImageMetadataDao(connector)
    try:
        logging.info("Starting generator of ImageMetadata")
        cursor = im_dao.collection.find(
            query,
            {
                "_id": True,
                "_metadata": True,
                "dicom_tags": {"Not used": "Not used"},
                "file_source": True,
                "study_id": True,
                "series_id": True,
                "aka_file_sources": True,
            },
        )
        for x in cursor:
            yield x
    finally:
        cursor.close()


def im_meta_from_virtual_dataset(
    ds: pydicom.Dataset, file_source: FileSource, _id: ObjectId = None
):
    """An alternative version of ImageMetadata.from_dataset that does not require a
    ds.filename

    :param ds: Dicom dataset
    :param file_source: Source of the dataset
    :param _id: the id that will be stored in the dst_mongo
    """
    json_dict = ds.to_json_dict(suppress_invalid_tags=True)
    dicom_tags = DatasetMod.tags_to_keywords(json_dict)
    return ImageMetadata(
        _id=_id,
        series_id=None,
        _metadata=None,
        file_source=file_source,
        dicom_tags=dicom_tags,
    )


def process_batch(args):
    im_metas, cfg = args

    src_im_dao = ImageMetadataDao(Connector(**cfg.src_mongo))
    dst_im_dao = ImageMetadataDao(Connector(**cfg.dst_mongo))
    try:
        processed = 0
        seen = 0
        for im_meta in im_metas:
            seen += 1
            try:
                processed += process_im_meta(im_meta, cfg, src_im_dao, dst_im_dao)
            except IndexError:
                pass
    finally:
        src_im_dao.connector.close()
        dst_im_dao.connector.close()

    return seen, processed


def process_im_meta(
    im_meta_dict: dict,
    cfg,
    src_im_dao: ImageMetadataDao,
    dst_im_dao: ImageMetadataDao,
    retries=10,
):
    im_meta = ImageMetadata.from_dict(im_meta_dict)
    # get source file
    src_filepath = im_meta.get_local_filepath(cfg.mount_paths)
    try:
        src_ds = pydicom.read_file(src_filepath)
    except Exception as e:
        logging.error(f"Could not read {im_meta._id} - {e}")
        return 0

    # pseudonymize
    try:
        deid_ds = cfg.deider.pseudonymize(src_ds)
    except ValueError as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0
    except KeyError as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0
    except BytesLengthException as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0

    # get path where it will be saved
    dst_rel_dir = cfg.dst_rel_dir
    if dst_rel_dir is None:
        dst_rel_dir = ""
    dst_rel_filepath = os.path.join(dst_rel_dir, get_instance_path(deid_ds))
    dst_file_source = FileSource(cfg.dst_drive_name, dst_rel_filepath)
    if dst_file_source.drive_name in [
        x.get("drive_name") for x in im_meta_dict.get("aka_file_sources", [])
    ]:
        logging.debug("This file has already been created.")
        raise IndexError("This file has already been created.")
    dst_filepath = dst_file_source.get_local_filepath(cfg.mount_paths)

    os.makedirs(os.path.dirname(dst_filepath), exist_ok=True)
    try:
        pydicom.write_file(dst_filepath, deid_ds)
    except ValueError as e:
        logging.error(f"{e}")
    except OSError as e:
        print(f"{e} - {dst_filepath}")
        if retries > 0:
            sleep(2)
            print(f"Retrying now, {retries} retries left - {dst_filepath}")
            r = retries - 1
            return process_im_meta(im_meta_dict, cfg, src_im_dao, dst_im_dao, r)
        print(
            f"Unknown OS error, could not fix by retrying, will not write - {dst_filepath}"
        )

    # the _id of the dataset is stored with the same database _id so that a super user
    # with access to both collections, can quickly check that everything is okay.
    try:
        del deid_ds.PixelData
    except AttributeError as e:
        logging.error(f"{e}")

    deid_im_meta = im_meta_from_virtual_dataset(deid_ds, dst_file_source, im_meta._id)

    # Keep record of what happened
    try:
        dst_im_dao.insert_one(deid_im_meta)
    except DuplicateKeyError as e:
        logging.error(f"{e}")
        return 0
    src_im_dao.add_aka(im_meta._id, dst_file_source)
    return 1


def get_required_drive_names(src_mongo, query):
    connector = Connector(**src_mongo)
    im_dao = ImageMetadataDao(connector)
    logging.info("Searching images to obtain required drives for query")
    print(f"Running query: {query}")
    cursor = im_dao.collection.aggregate(
        [{"$match": query}, {"$sortByCount": "$file_source.drive_name"}]
    )
    drive_names = []
    images = 0
    for doc in cursor:
        drive_names.append(doc["_id"])
        images += doc["count"]
    return drive_names, images


def check_mount_paths(mount_paths, required_drive_names, dst_drive_name):
    missing = set(required_drive_names) - set(mount_paths.keys())
    assert missing == set(), f"Missing configuration for drive_names: {missing}"
    assert mount_paths.get(dst_drive_name) is not None, "Missing dst_drive_name"


def single_proc(batches, n_proc=1):
    for batch in batches:
        seen, processed = process_batch(batch)
        yield seen, processed


def multi_proc(batches, n_proc):
    pool = Pool(n_proc)
    for seen, processed in pool.imap_unordered(process_batch, batches):
        yield seen, processed


def query_mux(query, image_ids: List[ObjectId] = None):
    """Select one of query or image_ids"""
    if image_ids is not None:
        assert query == {}, "Cannot have a query and a list of images at the same time"
        query = {"_id": {"$in": image_ids}}
    return query


def recipe_mux(base: bool, mr: bool, recipe: Union[str, List[str]]):
    """Select the correct recipe according to the configurations"""
    recipes = []
    if base:
        print(f"Using VAIB recipe from: {base_recipe_path}")
        recipes.append(base_recipe_path)
    if mr:
        print(
            f"Using additional to VAIB recipe targeted to MR studies from: {mr_recipe_path}"
        )
        recipes.append(mr_recipe_path)
    if type(recipe) == str:
        recipes.append(recipe)
    if type(recipe) == list:
        recipes = recipes + recipe
    for r in recipes:
        assert os.path.exists(r), f"Deid recipe path does not exist: {r}"
    if len(recipes) == 0:
        return None
    logging.info(f"Using recipes: {recipes}")
    return recipes


def main(args=None, image_ids: List[ObjectId] = None):
    if args is None:
        # Allows testing main from pytest.
        args = sys.argv[1:]

    # read config file
    cfg = parse_arguments(args)
    cfg.query = query_mux(cfg.query, image_ids)

    # Check how many files will be processed and if the cfg is enough
    required_drive_names, total_imgs = get_required_drive_names(
        cfg.src_mongo, cfg.query
    )
    check_mount_paths(cfg.mount_paths, required_drive_names, cfg.dst_drive_name)
    im_meta_gen = im_meta_generator(cfg.src_mongo, cfg.query)
    logging.info("Saving list of files to be processed")
    imgs = list(
        tqdm(im_meta_gen, desc="Reading list of files to be processed")
    )  # save as list to avoid losing the cursor
    batches = batcher(imgs, batch_size=cfg.batch_size)

    # Due to a weird configuration system of the deid library this has to be loaded
    # after setting the MESSAGELEVEL env var.
    os.environ["MESSAGELEVEL"] = cfg.logging_level

    cfg.deider = Deider(
        cfg.hash_secret,
        recipe_mux(
            cfg.deid_default_recipes["base"],
            cfg.deid_default_recipes["mr"],
            cfg.user_recipe_path,
        ),
        logging_level=cfg.logging_level,
    )

    batches = add_args_to_iterable(batches, cfg)
    pbar_seen = tqdm(total=total_imgs, desc="Images seen", smoothing=0.001)
    pbar_processed = tqdm(total=total_imgs, desc="Images processed", smoothing=0.001)
    if cfg.n_proc == 1:
        process_func = single_proc
    if cfg.n_proc > 1:
        process_func = multi_proc
    if cfg.n_proc < 1:
        raise ValueError("n_proc must be bigger than 0")
    for seen, processed in process_func(batches, cfg.n_proc):
        pbar_processed.update(processed)
        pbar_seen.update(seen)



from multiprocessing import Pool
from typing import List, Tuple

from pydicom import Dataset
from tqdm import tqdm

from cobra_db.dataset_mod import DatasetMod
from cobra_db.model import Patient
from cobra_db.mongo_dao import Connector, PatientDao, StudyDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher
from cobra_db.utils import intersect_dicts_allow_empty_minority


def proto_patient_generator(connector_kwargs):
    pipeline = [
        {
            "$group": {
                "_id": {"$first": "$dicom_tags.PatientID.Value"},
                "study_ids": {"$push": "$_id"},
            }
        }
    ]

    connector = Connector.get_env_pass(**connector_kwargs)
    study_dao = StudyDao(connector)
    study_cursor = study_dao.collection.aggregate(pipeline, allowDiskUse=True)
    for i in study_cursor:
        yield i


def process_patient_batch(args: Tuple[List[dict], dict]):
    batch, connector_kwargs = args
    connector = Connector.get_env_pass(**connector_kwargs)
    study_dao = StudyDao(connector)
    patient_dao = PatientDao(connector)

    counter = 0

    for patient in batch:
        studies = [study_dao.get_by_id(i, obj=False) for i in patient["study_ids"]]
        dicom_tags = [s["dicom_tags"] for s in studies]
        intersection = intersect_dicts_allow_empty_minority(dicom_tags)
        shared_ds_dict = DatasetMod.keywords_to_tags(intersection)
        shared_ds = Dataset.from_json(shared_ds_dict)
        patient = Patient.from_dataset(shared_ds)
        patient_id = patient_dao.insert_one(patient)
        counter += 1
        for study in studies:
            study_dao.update_patient_id(study["_id"], patient_id)
    return counter


def main(connector_kwargs):
    batch_size = 5
    iterable = batcher(proto_patient_generator(connector_kwargs), batch_size)
    iterable = add_args_to_iterable(iterable, connector_kwargs)
    pbar = tqdm(desc="patients", smoothing=0.005)
    for batch in iterable:
        counter = process_patient_batch(batch)
        pbar.update(counter)


def main_multiproc(connector_kwargs, n_proc):
    pool = Pool(n_proc)
    iterable = batcher(proto_patient_generator(connector_kwargs), 5)
    iterable = add_args_to_iterable(iterable, connector_kwargs)
    pbar = tqdm(desc="patients", smoothing=0.005)
    for n in pool.imap_unordered(process_patient_batch, iterable, chunksize=1):
        pbar.update(n)



import logging
import os
from copy import copy
from dataclasses import dataclass, fields, is_dataclass
from datetime import datetime, timezone
from typing import Callable, List, Literal, Tuple, Type, Union
from pathlib import Path
import numpy as np
import pydicom
from bson import ObjectId
from pydicom.dataset import Dataset

from cobra_db import __version__
from cobra_db.dataset_mod import DatasetMod
from cobra_db.enums import Modality, PregnancyStatus
from cobra_db.utils import (
    intersect_dicts,
    parse_AS_as_int,
    parse_DA_as_datetime,
    parse_DA_TM_as_datetime,
)


class BaseObj:
    """Defines the interface to load the object from dict and how to export it to dict."""

    @staticmethod
    def expand(value):
        if isinstance(value, BaseObj):
            return value.to_dict()
        else:
            return value

    def _to_dict(self):
        if is_dataclass(self):
            return self.__dict__
        else:
            raise NotImplementedError

    def to_dict(self):
        """expand the nested objects and remove all keys with None value"""
        d = {k: self.expand(v) for k, v in self._to_dict().items() if v is not None}
        return d

    @classmethod
    def from_dict(cls, d: dict):
        raise NotImplementedError(
            f"This method should be implemented in the child {type(cls)}"
        )


class Embedded(BaseObj):
    """Base class for an embedded doc (that doesn't get its own _id)"""

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)


@dataclass
class Metadata(Embedded):
    """Stores information of when the db doc was created, modified, and the data model
    version.
    Should be used in any doc with _id that is stored in the database.
    """

    model_version: str
    created: datetime
    modified: datetime = None
    project_name: str = None

    @classmethod
    def create(cls):
        return cls(model_version=__version__, created=datetime.now(timezone.utc))


@dataclass
class Entity(BaseObj):
    """Base class for all entities that are stored in the db with an _id, i.e. that
    require a Dao.
    Usually parent of a ``@dataclass`` class
    """

    _id: ObjectId
    _metadata: Metadata

    def __repr__(self):
        header = f"<{self.__class__.__name__}>"
        body = copy(self.__dict__)
        return f"{header}\n{body}\n"

    def __post_init__(self):
        """Automatically add db_metadata if missing"""
        if self._metadata is None:
            self._metadata = Metadata.create()
        if isinstance(self._metadata, dict):
            self._metadata = Metadata(**self._metadata)

    @classmethod
    def from_dict(cls, obj_dict: dict):
        obj = cls(**obj_dict)
        for f in fields(obj):
            typ = f.type
            if type(typ) != type:
                # TODO: not handled yet!!!
                pass
            else:
                val_dict = obj_dict.get(f.name, None)
                if val_dict is not None and issubclass(typ, BaseObj):
                    obj.__setattr__(f.name, typ.from_dict(val_dict))
        return obj


class Source:
    """Base class for a pointer to more information."""


@dataclass
class DicomEntity(Entity):
    """Base class for all entities that contain dicom_tags."""

    dicom_tags: dict

    @staticmethod
    def optional(ds: Dataset, tag_name: str, parser: Callable = None, default=None):
        """
        Get tag value from dataset, if missing return default value.
        """
        value = ds.get(tag_name, None)
        if value is not None and value != "":
            if parser is not None:
                value = parser(value)
            return value
        else:
            return default

    @staticmethod
    def choose_first(ds, *tag_names, parser, default):
        """
        Choose first tag that does not return None,
        """
        for tag_name in tag_names:
            tag = DicomEntity.optional(ds, tag_name, parser, None)
            if tag is not None:
                return tag
        return default

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """Abstract method. How to create an instance from a pydicom Dataset"""
        raise NotImplementedError(f"Should be implemented by child {type(cls)}")

    def get_tag(self, keyword: str, default=None):
        tag = self.dicom_tags.get(keyword, None)
        if tag is not None:
            value = tag.get("Value", default)
            if not value == default:
                if isinstance(value, list):
                    if len(value) == 1:
                        return value[0]  # most of the cases are this
                return value
        else:
            return default

    def get_acquisition_datetime(self):
        time = self.get_tag("AcquisitionTime")
        time = "000000" if time is None else time
        date = self.get_tag("AcquisitionDate")
        date = "00000000" if date is None else date
        return parse_DA_TM_as_datetime(date, time)


@dataclass
class EntitySource(Embedded, Source):
    """Pointer to another Entity instance in the database.
    Should only be used when there's ambiguity about the referenced class/collection.
    """

    name: str
    description: str
    _entity_type: Union[Type[Entity], str]
    _id: ObjectId

    def __post_init__(self):
        if not type(self._entity_type) is str:
            self._entity_type = self._entity_type.__name__

    def to_dict(self):
        return self.__dict__


@dataclass
class FileSource(Embedded, Source):
    """Pointer to a file in the filesystem.
    rel_path is relative to the place where drive_name is mounted.
    This is because drive_name can be mounted in different paths according
    to the machine and we don't have another way of referencing a drive.
    filename is automatically stored for easy indexing and querying without having to
    manipulate the rel_path to extract the filename.
    """

    drive_name: str
    rel_path: str
    filename: str = None  # automatically set when rel_path is set in __post_init__

    def get_local_filepath(self, mount_paths: dict) -> str:
        """Returns the filepath according to mount_paths"""
        return os.path.join(mount_paths[self.drive_name], self.rel_path)

    def __post_init__(self):
        if self.filename is None:
            self.filename = os.path.basename(self.rel_path)

    @classmethod
    def from_mount_paths(cls, filepath: str, mount_paths: dict):
        """create an instance from the filepath and the mount_paths.

        :param filepath: _description_
        :param mount_paths: _description_
        """
        for drive_name, mount_path in mount_paths.items():
            path = Path(filepath)
            mount_path = Path(mount_path)
            rel_path = str(path.relative_to(mount_path).as_posix())
            return cls(drive_name=drive_name, rel_path=rel_path)
        raise ValueError(f"{filepath} is not in any of the mount paths: {mount_paths}")


@dataclass
class Patient(Entity):
    """Represents a unique person that has gone through at least one study.
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.2.2.html
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.2.3.html
    """

    anon_id: str
    hidden: bool
    birth_date: datetime = None

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """
        Create an instance of the entity from a DICOM dataset
        """
        return cls(
            _id=None,
            _metadata=None,
            anon_id=ds.PatientID,
            birth_date=DicomEntity.optional(
                ds, "PatientBirthDate", parse_DA_as_datetime, None
            ),
            hidden=False,
        )

    def __eq__(self, other: "Patient") -> bool:
        return self.anon_id == other.anon_id


@dataclass
class RadiologicalStudy(DicomEntity):
    """Aggregation of multiple dicom instances by the combination of patient and date."""

    study_uid: List[str]  # (0020,0010) LO StudyInstanceUID
    date: datetime  # (0008,0020) DA StudyDate + (0008,0030) TM StudyTime
    series_count: int  # (0020,0011) IS SeriesNumber, Overriden after grouping because
    # the tags are not reliable.
    patient_id: ObjectId = None
    # SOPClassUID, the meaning can be found in pydicom._uid_dict.UID_dictionaty

    # Tags that we want to keep even if the images in the same patient+date disagree.

    sop_class: List[str] = None
    modality: List[Modality] = None  # (0008,0060) CS Modality
    description: str = None  # (0008,1030) LO StudyDescription
    accession_number: List[
        str
    ] = None  # Used to link the study with other administrative registers.
    patient_age: int = None  # in years (0010,1010) AS PatientAge
    patient_weight: float = None  # (0010,1030) PatientWeight
    patient_height: float = None  # (0010,1020) PatientSize
    patient_pregnancy_status: PregnancyStatus = None  # (0010,21C0)
    manufacturer: List[str] = None
    manufacturer_model_name: List[str] = None
    detector_id: List[str] = None
    detector_type: List[str] = None
    device_serial_number: List[str] = None
    software_versions: List[str] = None
    date_of_last_detector_calibration: List[str] = None
    breast_implant_present: List[str] = None

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """Create an instance of the entity from a DICOM dataset"""
        try:
            modality = [cls.optional(ds, "Modality", Modality, None)]
        except ValueError as e:
            logging.error(f"{e}")
            modality = None

        return cls(
            _id=None,
            _metadata=None,
            study_uid=cls.optional(ds, "StudyInstanceUID", default=None),
            modality=modality,
            sop_class=[cls.optional(ds, "SOPClassUID", str, None)],
            date=parse_DA_TM_as_datetime(
                cls.optional(ds, "StudyDate", str, "00000000"),
                cls.optional(ds, "StudyTime", str, "000000"),
            ),
            description=cls.optional(ds, "StudyDescription"),
            accession_number=cls.optional(ds, "AccessionNumber", str, None),
            patient_age=cls.optional(ds, "PatientAge", parse_AS_as_int, None),
            patient_weight=cls.optional(ds, "PatientWeight", float, None),
            patient_height=cls.optional(ds, "PatientSize", float, None),
            patient_pregnancy_status=(
                cls.optional(ds, "PatientPregnancyStatus", PregnancyStatus, None)
            ),
            dicom_tags=DatasetMod.tags_to_keywords(ds.to_json_dict()),
            # The ones below all get overriden while grouping
            series_count=cls.optional(ds, "SeriesNumber", int, None),
            manufacturer=cls.optional(ds, "Manufacturer", str, None),
            manufacturer_model_name=cls.optional(
                ds, "ManufacturerModelName", str, None
            ),
            detector_id=cls.optional(ds, "DetectorID", str, None),
            detector_type=cls.optional(ds, "DetectorType", str, None),
            device_serial_number=cls.optional(ds, "DeviceSerialNumber", str, None),
            software_versions=cls.optional(ds, "SoftwareVersions", None, None),
            date_of_last_detector_calibration=cls.optional(
                ds, "DateOfLastDetectorCalibration", parse_DA_as_datetime, None
            ),
            breast_implant_present=cls.optional(ds, "BreastImplantPresent", str, None),
        )

    @classmethod
    def from_dict(cls, d: dict):
        """Convert a mongodb dict into a RadiologicalStudy object"""
        return cls(**d)

    def __eq__(self, other: "RadiologicalStudy") -> bool:
        return self.anon_study_id == other.anon_study_id


@dataclass
class RadiologicalSeries(DicomEntity):
    study_id: ObjectId
    date: datetime = None  # (0008,0021) DA SeriesDate + (0008,0031) TM SeriesTime
    series_uid: str = None  # (0020,000E) UI SeriesInstanceUID
    description: str = None  # 0008103E
    # Overrided by the count in the ImageMetadata collection
    image_count: int = None  # (0020,0013) IS InstanceNumber.
    image_shape: Tuple[int, int] = None  # If the dicom_tags of all images agree.
    protocol_name: str = None  # (0018,1030) LO ProtocolName,
    # other possible tags 00180018, 0008103E, 00181081, 00181250, 00180024, 00181030

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """
        Create an instance of the entity from a DICOM dataset
        """
        return cls(
            _id=None,
            _metadata=None,
            series_uid=cls.optional(ds, "SeriesInstanceUID", str, None),
            date=parse_DA_TM_as_datetime(
                cls.optional(ds, "SeriesDate", str, "00000000"),
                cls.optional(ds, "SeriesTime", str, "000000"),
            ),
            study_id=None,
            image_count=cls.optional(ds, "InstanceNumber", int, None),
            protocol_name=cls.optional(ds, "ProtocolName", str, None),
            description=cls.optional(ds, "SeriesDescription", str, None),
            image_shape=(
                cls.optional(ds, "Rows", int, None),
                cls.optional(ds, "Columns", int, None),
            ),
            dicom_tags=DatasetMod.tags_to_keywords(ds.to_json_dict()),
        )

    @classmethod
    def from_images_metadata(cls, images_metadata: List["ImageMetadata"]):
        #  Shared properties are kept
        shared_dicom_tags = intersect_dicts(
            [im_meta.dicom_tags for im_meta in images_metadata]
        )
        _shared_dicom_dict = DatasetMod.keywords_to_tags(
            shared_dicom_tags
        )  # back to dicom hex tags
        shared_ds = Dataset.from_json(_shared_dicom_dict)
        return cls.from_dataset(shared_ds)

    def __eq__(self, other: "RadiologicalSeries") -> bool:
        return self.series_uid == other.series_uid


@dataclass
class ImageMetadata(DicomEntity):
    """Represents one DICOM image i.e. a DICOM instance with PixelData.
    Only the dicom headers are stored. While the pixel data is not stored.
    """

    dicom_tags: dict
    file_source: FileSource
    study_id: ObjectId = None
    series_id: ObjectId = (
        None  # is added by the method ImageMetadataDao.update_series_id()
    )
    aka_file_sources: List[FileSource] = None

    @classmethod
    def from_dataset(cls, ds: Dataset, mount_paths: dict):
        """Create one instance from a pydicom Dataset.

        :param ds: pydicom Dataset
        :param mount_paths: dict that maps the drive_names to the mount path of the
         drives that are in use.
        """
        # sometimes (<1%) the dataset contains mismatches on the VR. For example IS with
        # a float value.
        # in those cases the tag has to be fixed to be able to create the dict.
        json_dict = ds.to_json_dict(suppress_invalid_tags=True)
        dicom_tags = DatasetMod.tags_to_keywords(json_dict)
        return cls(
            _id=None,
            series_id=None,
            _metadata=None,
            file_source=FileSource.from_mount_paths(ds.filename, mount_paths),
            dicom_tags=dicom_tags,
        )

    def to_dataset(self) -> Dataset:
        """Creates a pydicom Dataset with only the metadata in dicom_tags.
        Useful to use pydicom interface when required.
        """
        dicom_tags = DatasetMod.keywords_to_tags(self.dicom_tags)
        return Dataset.from_json(dicom_tags)

    def get_checksum(self, mount_paths: dict) -> str:
        """Obtain the md5 hexdigest hash of the original file. This helps finding
        duplicated images.
        Warning: md5 is not cryptographycally secure.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: string hexdigest with the hash of the file.
        """
        import hashlib

        filepath = self.get_local_filepath(mount_paths)
        with open(filepath, "rb") as file:
            return hashlib.md5(file.read()).hexdigest()

    def get_local_filepath(self, mount_paths: dict) -> str:
        """Returns the filepath of the file_source according to mount_paths.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: string hexdigest with the hash of the file.
        """
        return self.file_source.get_local_filepath(mount_paths)

    def get_pixel_data(self, mount_paths) -> np.ndarray:
        """Load the file and extract the pixel_array from the pydicom dataset.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: the content of pydicom.Dataset.pixel_array
        """
        filepath = self.get_local_filepath(mount_paths)
        ds = pydicom.dcmread(filepath)
        return ds.pixel_array

    def get_file_size(self, mount_paths) -> int:
        """Get the size in bytes of the file"""
        filepath = self.get_local_filepath(mount_paths)
        return os.path.getsize(filepath)

    def __repr__(self):
        return super().__repr__()


@dataclass
class Annotation(Entity):
    """
    Abstract class for any annotation that is stored in cobra_db. Should be subclassed
    """

    annotation_type: List[
        Literal["text", "semantic_segmentation", "binary", "categorical"]
    ]
    annotator_email: str
    date: datetime  # annotation date
    possible_labels: List[
        str
    ]  # List of all possible labels example [mass, non-mass, ...]
    file_format: str  # Name of the extension of the file. jpeg, png, npz, etc.
    aggregation_level: Literal["patient", "study", "series", "image"]
    # For example the SOPInstanceUID of the image, StudyInstanceID of the study, etc.
    referenced_entity_uid: str
    patient_anon_id: str
    study_date: datetime
    file_source: FileSource  # file/folder of the annotation
    project_name: str


@dataclass
class Annotator(Entity):
    email: str
    specialization: str
    first_name: str
    last_name: str



import re
from datetime import datetime
from typing import Tuple

from Crypto.Hash import SHA512

from cobra_db.enums import PersonalNumberValidity


class PersonalNumberValidator:
    """Class to group all methods related to validating personal numbers.
    Checksum is calculated according to
    https://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)#Checksum
    """

    @staticmethod
    def get_checksum(personal_number: str) -> int:
        """Calculate the checksum for a personal number

        :param personal_number: 10 digit pnr. Without any other characters.
        """
        identity = "212121212"
        multiply = [str(int(p) * int(i)) for p, i in zip(personal_number, identity)]
        sum_digits = str(sum(int(m) for m in "".join(multiply)))
        last_digit = sum_digits[-1]
        return 10 - int(last_digit)

    @staticmethod
    def validate(personal_number: str) -> PersonalNumberValidity:
        """
        Checks different aspects of validity described in the PnrValidity enum.
        It is important to store this information along with the hashed PersonalNumber
        to be able to debug matching problems later. This is because the hashed version
        of 19000101-1234, 190001011234 and 0001011234 are totally different.
        At least with this information we can try to modify one side or the other of the
        matching to fix it.
        """

        validity = PersonalNumberValidity.NOT_VALID
        # check for characters outside 0-9
        p = personal_number
        personal_number = re.sub(r"[^\da-zA-Z]+", "", personal_number)
        if p == personal_number:
            validity = validity | PersonalNumberValidity.CHARS

        if len(personal_number) == 10:
            pass
        elif len(personal_number) == 12:  # long version of the personnummer
            validity = validity | PersonalNumberValidity.LENGTH
            # make it short since the other two digits are not used in the checksum
            personal_number = personal_number[2:]
        else:
            return validity  # must be either 10 or 12 digit to continue

        try:  # parse the date to and raise errors if something wrong
            datetime.strptime(personal_number[:6], "%y%m%d")
            validity = validity | PersonalNumberValidity.DATE
        except ValueError:
            pass  # the date could not be parsed

        try:  # checksum
            checksum = PersonalNumberValidator.get_checksum(personal_number)
            if int(personal_number[-1]) == checksum:
                validity = validity | PersonalNumberValidity.CHECKSUM
        except ValueError:
            pass  # could not compute checksum
        except IndexError:
            pass

        return validity


class Hasher:
    """Hash method for the VAIB project, uses a sha512/256 hashing which is
    "a one way cryptographic process", i.e. irreversible.
    """

    def __init__(self, secret_salt: str) -> None:
        """Create an instance of the Hasher that has a predefined salt

        :param secret_salt: a random string that makes the hash harder to break. It gets
        prepended to the string that gets hashed.
        """
        # important because when getting the bytes anything can be used as salt.
        assert type(secret_salt) == str
        self.secret_salt = secret_salt

    def hash(self, msg: str):
        """Hash a message

        :param msg: message that we want to encrypt, normally the personnummer or the
         StudyID.
        :return: the encrypted message as hexdigest (in characters from '0' to '9' and
         'a' to 'f')
        """
        assert type(msg) == str, f"value is not of type str, {type(msg)}"
        h = SHA512.new(truncate="256")
        bytes_str = bytes(f"{self.secret_salt}{msg}", "utf-8")
        h.update(bytes_str)
        return str(h.hexdigest())

    def hash_personal_number(
        self, personal_number: str
    ) -> Tuple[str, PersonalNumberValidity]:
        """Hash the personnal number after removing all characters that are not 0-9,
        a-z, or A-Z.
        Obtain the validity of the personal number before the hashing.

        :param personal_number: personal number to be hashed
        :return: hashed string and validity of the pre-hashed pnr
        """
        pnr = re.sub(r"[^\da-zA-Z]+", "", personal_number)
        return self.hash(pnr), PersonalNumberValidator.validate(pnr)



# In stage 2 we create the hierarchy of files from the metadata.
# First, we group files by PatientID+StudyDate

import logging
from multiprocessing import Pool
from typing import List, Tuple

from pydicom import Dataset
from tqdm import tqdm

from cobra_db.dataset_mod import DatasetMod
from cobra_db.model import RadiologicalStudy
from cobra_db.mongo_dao import Connector, ImageMetadataDao, SeriesDao, StudyDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher
from cobra_db.utils import intersect_dicts_allow_empty_minority


def proto_study_generator(connector_kwargs):
    query = [
        {
            "$group": {
                "_id": {
                    "patient_id": {"$first": "$dicom_tags.PatientID.Value"},
                    "study_date": {"$first": "$dicom_tags.StudyDate.Value"},
                },
                "image_ids": {"$push": "$_id"},
            }
        }
    ]
    connector = Connector.get_env_pass(**connector_kwargs)
    im_dao = ImageMetadataDao(connector)
    proto_study_cursor = im_dao.collection.aggregate(query, allowDiskUse=True)
    # Store it in a list otherwise the cursor might die after ~30min
    proto_study_list = list(proto_study_cursor)
    for i in proto_study_list:
        yield i


def get_tag(d, keyword):
    try:
        v = d[keyword]["Value"]
        if isinstance(v, list):
            return set(v)
        else:
            return set([v])
    except KeyError:
        return set()


def set_union(dicom_tags: List[dict], keyword):
    values = set()
    for d in dicom_tags:
        values = values.union(get_tag(d, keyword))
    if len(values) == 0:
        return None
    return list(values)


def group_study(images: List[dict], project_name: str) -> RadiologicalStudy:
    """Takes the dict of many ImageMetadata instances.
    Obtains the shared dicom_tags and extract more information to put it in the root
    of the document.

    :param images: List of dictionaries from the ImageMetadata collection
    :param project_name: Project to which this study belongs.
    """
    dicom_tags = [i["dicom_tags"] for i in images]
    intersection = intersect_dicts_allow_empty_minority(dicom_tags)
    shared_ds_dict = DatasetMod.keywords_to_tags(intersection)
    shared_ds = Dataset.from_json(shared_ds_dict)
    try:
        study = RadiologicalStudy.from_dataset(shared_ds)
    except ValueError as e:
        logging.error(
            f"Could not create study instance for \
    {get_tag(dicom_tags[0], 'PatientID')}-{get_tag(dicom_tags[0], 'StudyDate')}, e:{e}"
        )
        return None

    union_keywords = {
        "study_uid": "StudyInstanceUID",
        "accession_number": "AccessionNumber",
        "modality": "Modality",
        "sop_class": "SOPClassUID",
        "manufacturer": "Manufacturer",
        "manufacturer_model_name": "ManufacturerModelName",
        "detector_id": "DetectorID",
        "detector_type": "DetectorType",
        "device_serial_number": "DeviceSerialNumber",
        "software_versions": "SoftwareVersions",
        "date_of_last_detector_calibration": "DateOfLastDetectorCalibration",
        "breast_implant_present": "BreastImplantPresent",
    }
    for attr_name, keyword in union_keywords.items():
        setattr(study, attr_name, set_union(dicom_tags, keyword))

    study.series_count = len(set_union(dicom_tags, "SeriesInstanceUID"))
    study._metadata.project_name = project_name
    return study


def process_study_batch(args: Tuple[List[dict], dict, str]):
    batch, connector_kwargs, project_name = args
    connector = Connector.get_env_pass(**connector_kwargs)
    im_dao = ImageMetadataDao(connector)
    study_dao = StudyDao(connector)
    series_dao = SeriesDao(connector)
    counter = 0
    for study in batch:
        images = [im_dao.get_by_id(i, obj=False) for i in study["image_ids"]]
        study = group_study(images, project_name)
        if study is None:
            continue
        study_id = study_dao.insert_one(study)
        updated_series = set()
        for i in images:
            im_dao.update_study_id(i["_id"], study_id)
            series_id = i.get("series_id")
            if series_id and series_id not in updated_series:
                series_dao.update_study_id(series_id, study_id)
                updated_series.add(series_id)
        counter += 1
    return counter


def main(connector_kwargs, project_name):
    iterable = batcher(proto_study_generator(connector_kwargs), 5)
    iterable = add_args_to_iterable(iterable, connector_kwargs, project_name)
    pbar = tqdm(desc="studies", smoothing=0.05)
    for batch in iterable:
        n = process_study_batch(batch)
        pbar.update(n)


def main_multiproc(connector_kwargs, n_proc, project):
    pool = Pool(n_proc)
    iterable = batcher(proto_study_generator(connector_kwargs), 5)
    iterable = add_args_to_iterable(iterable, connector_kwargs, project)
    pbar = tqdm(desc="studies", smoothing=0.005)
    for n in pool.imap_unordered(process_study_batch, iterable, chunksize=1):
        pbar.update(n)
